//
//  covalbridge.h
//
//  coval version 0.81
//      2012.09.30 Takeshi Ogihara
//  Copyright (C) 2010, 2011, 2012, 2013 Takeshi Ogihara
//
#ifndef __COVAL_BRIDGE_H__
#define __COVAL_BRIDGE_H__

#include "coval.h"

#ifndef USE_LONGLONG
#define USE_LONGLONG    1
#endif
#ifndef USE_LONGDOUBLE
#define USE_LONGDOUBLE  0
#endif

/* Coval Bridge */
typedef int (*coval_bridge_convert)(void *to, const void *frm);
/* Conversion function */

#define _CONV_FromMASK  0x03
#define _CONV_ToMASK    0x0c
#define _CONV_OpeMASK   0xf0
#define Check_mask      0x0f
#define Func_mask  _CONV_OpeMASK // 0xf0
#define _CONV_ToShift   2
#define BR_NONE         0
#define BR_INT          1
#define BR_UINT         2
#define BR_FLT          3
#define BR_IntToInt     (BR_INT | (BR_INT << _CONV_ToShift))
#define BR_IntToFloat   (BR_INT | (BR_FLT << _CONV_ToShift))
#define BR_UIntToInt    (BR_UINT | (BR_INT << _CONV_ToShift))
#define BR_UIntToFloat  (BR_UINT | (BR_FLT << _CONV_ToShift))
#define BR_FloatToInt   (BR_FLT | (BR_INT << _CONV_ToShift))
#define BR_FloatToFloat (BR_FLT | (BR_FLT << _CONV_ToShift))
#define _BR_Conv        0x10
#define BR_NumNeg       0x20
#define BR_LogicNeg     0x30 // "to" should be int
#define BR_CopyBytes    0x40

typedef struct _bridge_base {
    coval_base *_cv[2];
    unsigned char _sw[2];   // check | Func/Ope.
    unsigned short _sz[2];  // size of data
    union { coval_bridge_convert func; unsigned char conv[2]; } _fn[2];
    coval_base bank[2];
} bridge_base;
// conv[2]: from-type, to-type
// if (check && (assigned value == the previous)), don't activate callbacks.

#define CovalBridgeOf(TYP,TYQ)  struct { coval_base *_cv[2]; \
  unsigned char _sw[2]; unsigned short _sz[2]; \
  union { coval_bridge_convert func; unsigned char conv[2]; } _fn[2]; \
  CovalTypeOf(TYP) bank1; CovalTypeOf(TYQ) bank2; }
void _coval_bridge_initialize(bridge_base *, unsigned, unsigned);
#define CovalBridgeInit(TYP, TYQ, BRp) do { \
  typedef CovalBridgeOf(TYP,TYQ) _ty; _ty *p = (_ty *)(BRp); \
  p->_cv[0] = (coval_base *)&p->bank1; p->_cv[1] = (coval_base *)&p->bank2; \
  _coval_bridge_initialize((bridge_base *)(BRp), sizeof(TYP), sizeof(TYQ)); \
}while(0)

#define CovalBridge(BRp, X, TARGp)  _bindCoval((coval_base *)(TARGp), \
  ((bridge_base *)(BRp))->_cv[X], NULL)
#define BridgeCoval     CovalBridge       // for compatibility (0.5)

#define CovalBridgeMember(BRp, X, TARGp, M) _bindCoval( \
  (coval_base *)(TARGp), \
  ((bridge_base *)(BRp))->_cv[X], &(TARGp)->_p->M)
#define BridgeCovalMember       CovalBridgeMember   // for compatibility (0.5)

#define CovalUnbridge(BRp, X)   UnbindCoval((BRp)->_cv[X])
#define UnbridgeCoval   CovalUnbridge   // for compatibility (0.5)

int _setCovalBridgeConversion(bridge_base *brg, int idx, int conv, int check);
#define CovalSetBridgeConversion(BRp, X, CONV, CK) \
  _setCovalBridgeConversion((bridge_base *)(BRp), (X), (CONV), (CK))
#define CovalSetBridgeFunction(BRp, X, FN, CK) \
  (BRp)->_sw[X] = (Func_mask | ((CK)?Check_mask:0)), \
  (BRp)->_fn[X].func = (coval_bridge_convert)(FN)
#define SetCovalBridgeConversion    CovalSetBridgeConversion
    // for compatibility (0.5)
#define SetCovalBridgeFunction   CovalSetBridgeFunction
    // for compatibility (0.5)

#endif /* __COVAL_BRIDGE_H__ */
